﻿using System;

using Buddy.Wildstar.Engine;
using Buddy.Wildstar.Game;
using Buddy.Wildstar.Game.Actors;
using Buddy.Wildstar.BotCommon;

using log4net;
using MahApps.Metro.Controls;
using System.IO;
using System.Windows.Markup;
using System.Runtime.InteropServices;


namespace CombatBotEx {

    public class CombatBotEx : IBot, IUIButtonProvider {

        private ILog Log = LogManager.GetLogger("CombatBotEx");
        private static GlobalBotSettings AppSettings = new GlobalBotSettings();

        #region DLLImport to support Hotkeys

        // Source https://msdn.microsoft.com/en-us/library/windows/desktop/ms646301(v=vs.85).aspx
        [DllImport("user32.dll", CharSet = CharSet.Auto, ExactSpelling = true, CallingConvention = CallingConvention.Winapi)]
        public static extern short GetKeyState(int keyCode);

        [DllImport("user32.dll")]
        private static extern IntPtr GetForegroundWindow();

        //Modifier key vkCode constants - https://msdn.microsoft.com/en-us/library/windows/desktop/dd375731(v=vs.85).aspx
        private const int VK_SHIFT = 0x10;
        private const int VK_CONTROL = 0x11;
        private const int VK_MENU = 0x12;
        private const int VK_CAPITAL = 0x14;
        private const int VK_X = 0x58;
        
        #endregion

        #region IAuthored Implementation
        public string Author { get { return "Paulmuadip, Satbuster"; } }
        public string Name { get { return "CombatBotEx"; } }
        public string Version { get { return "1.0.6"; } }
        #endregion

        #region IBot Implementation

        private bool _pulseMe = false;
        private bool _paused = false;
        private bool _hotkeyDown = false;
        private bool _modifierDown = false;
        private bool _hotkeyPressed = false;
        
        public void OnRegistered() { }
        public void OnUnregistered() { }

        public void OnInitialize() {
            Log.Debug("OnInitialize");

            CombatBotExConfig.Instance.SettingsVersion = string.Format("Version {0} by {1}", this.Version, this.Author);
            _pulseMe = false;
            Log.Info("LastBot: " + CombatBotEx.AppSettings.LastBot + " LastCombatRoutine: " + CombatBotEx.AppSettings.LastCombatRoutine);
        }

        public void OnUninitialize() {
            Log.Debug("OnUninitialize");
            _pulseMe = false;
        }

        
        private void StartCombatRoutine() {
            GameEngine.BotPulsator.RegisterForPulse(GameEngine.CurrentRoutine);
        }

        public void OnStart() {
            Log.Info("Bot started.");
            _pulseMe = true;
            GameEngine.BotPulsator.RegisterForPulse(GameManager.Lua.Events);
            SpellManager.SetSpellCastingBehavior(SpellCastingBehavior.Normal);
            StartCombatRoutine();
        }

        private void StopCombatRoutine() {
            if (GameEngine.CurrentRoutine != null) {
                GameEngine.BotPulsator.UnregisterForPulse(GameEngine.CurrentRoutine);
            }
        }

        public void OnStop() {
            Log.Info("Bot stopped.");
            _pulseMe = false;
            GameEngine.BotPulsator.UnregisterForPulse(GameManager.Lua.Events);
            StopCombatRoutine();
        }


        public bool CanBePulsed { get { return _pulseMe; } }

        public bool Equals(IBot other) {
            if (this.Name != other.Name)
                return false;

            if (this.Version != other.Version)
                return false;

            if (this.Author != other.Author)
                return false;

            return true;
        }

        #endregion

        #region IBot : IPulsable.Pulse() Implementation

        // Todo: move to be configurable
        private int throttleDelay = 1; // delay in seconds for considering actions during Pulse()
        private int collectibleRadius = 5; // how far to look for collectibles
        private int maxLevelDifference = 3; // how many levels above player's level to auto-attack
        private double minPlayerHealthPercent = 0.5; // minimum player health percentage required for auto-attack

        private DateTime throttle = DateTime.Now;

        public void Pulse() {

            #region Check Hotkey
            // Check for hotkey presses, if game window is active and bot is running
            if ((GameManager.Memory.Process.MainWindowHandle == GetForegroundWindow()) && (_pulseMe == true) && CombatBotExConfig.Instance.SettingsHotkey)
                CheckHotkey();
            #endregion

            #region Throttle - Skip pulses
            // Only evaluate actions every 'throttleDelay' seconds.  Throttle activities.
            if ((DateTime.Now - this.throttle).TotalSeconds < this.throttleDelay) return;
            this.throttle = DateTime.Now;
            #endregion

            #region Actions
            // Try to loot collectibles or loot
            if (CombatBotExConfig.Instance.SettingsAutoLoot && !_paused) VacuumLoot();
            if (CombatBotExConfig.Instance.SettingsAutoCollect && !_paused) LootCollectibles();
            if (CombatBotExConfig.Instance.SettingsAutoAttack && !_paused) EngageTarget();
            #endregion
        }

        #endregion

        #region CheckHotkey Implementation

        public void CheckHotkey() {

            _hotkeyDown = (GetKeyState(VK_X) < 0);
            _modifierDown = (GetKeyState(VK_MENU) < 0);

            if (_hotkeyDown && _modifierDown && !_hotkeyPressed) {
                _hotkeyPressed = true;
                _paused = !_paused;
                if (_paused)
                    StopCombatRoutine();
                else
                    StartCombatRoutine();
                Log.Info("CombatBotX Paused? " + _paused);
            } else if ( !_hotkeyDown || !_modifierDown ) {
                _hotkeyPressed = false;
            }
        }

        #endregion

        #region EngageTarget Implementation

        private SpellManager.SpellInformation attackSpell = SpellManager.GetMiscSpell(MiscSpellType.RightClickAttack);

        public void EngageTarget() {

            // Make sure there's a valid me and a valid target
            Actor me = GameManager.LocalPlayer;
            if (me == null) return;
            Actor target = me.CurrentTarget;
            if (target == null) return;
            CreatureInfo info = target.CreatureInfo;
            if (info == null) return;

            if (CombatBotExConfig.Instance.SettingsCheckLevel) {
                // Don't auto-attack if target too high level
                int levelDifference = (int)target.Level - (int)me.Level; // positive implies target is of greater level than player
                if (levelDifference > maxLevelDifference) {
                    Log.Warn(info.Name + " [" + target.Level + "] is too high level. Not auto-engaging.");
                    return;
                }
            }

            if (CombatBotExConfig.Instance.SettingsCheckHealth) {
                // Abort engaging target if player health below minimum
                if (me.HealthPercent <= minPlayerHealthPercent)
                    return;
            }

            // Sanity check target
            if (target.IsAlive && (!me.IsInCombat) && (info.CanBeAttacked)) {
                Log.Info("Attacking " + info.Name + " with " + me.HealthPercent + "% health");
                if (CombatBotExConfig.Instance.SettingsAutoFace) target.Face();
                this.attackSpell.Cast();
            }
        }

        #endregion

        #region Settings Dialog Implementation and Handling

        public string ButtonText { get { return "BOT SETTINGS"; } }

        public System.Windows.Controls.UserControl Control {
            get {
                Log.Info("Loading XAML");
                using (var fs = new FileStream(@"Bots\CombatBotEx\CombatBotExGui.xaml", FileMode.Open)) {
                    var root = (System.Windows.Controls.UserControl)XamlReader.Load(fs);
                    root.DataContext = CombatBotExConfig.Instance; // WPF DataBinding, bind it to our CombatBotExConfig instance
                    return root;
                };
            }
        }

        public void OnButtonClicked(object sender) {
            Log.Info("Opening Configuration Dialog");
            MetroNavigationWindow configWindow = new MetroNavigationWindow() {
                Title = "CombatBotEx Settings",
                Content = this.Control,
                ShowCloseButton = false,
                ShowMinButton = false,
                ShowMaxRestoreButton = false,
                ShowIconOnTitleBar = false,
                SizeToContent = System.Windows.SizeToContent.WidthAndHeight
            };
            configWindow.ShowDialog();
            CombatBotExConfig.Instance.Save();
        }

        #endregion

        #region Looting Implementations

        public void LootCollectibles() {
            // Check if there are any collectibles (C) nearby
            Buddy.Common.DefaultDictionary<uint, Actor> objects = GameManager.Actors;
            foreach (var o in objects.Values) {
                if ((o.Distance < collectibleRadius) && (o.IsCollect)) {
                    Actor me = GameManager.LocalPlayer;
                    if ((!me.IsMoving) && (!me.IsOnTaxi) && (!me.IsBusy) && (!me.IsCasting)) {
                        Log.Info("Collecting " + o.Name);
                        o.Interact();
                    }
                }
            }
        }

        private SpellManager.SpellInformation lootSpell = SpellManager.GetMiscSpell(MiscSpellType.VacuumLoot);
        public void VacuumLoot() {
            // Check if there is anything to loot (V)
            if (GameManager.CanVacuum) {
                Log.Info("Looting");
                CommonBehaviors.VacuumLoot();
            }
        }

        #endregion
           
    }
}
